package com.loafle.commons.server;

import java.net.SocketAddress;
import java.util.List;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;

import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelOption;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.ServerChannel;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.handler.ssl.SslContext;

/**
 * Server
 */
public abstract class Server {
  @Value("${server.netty.thread.count.boss}")
  protected int threadCountBoss;
  @Value("${server.netty.thread.count.worker}")
  protected int threadCountWorker;

  @Autowired
  protected SslContext sslContext;
  @Autowired
  protected List<Class<ServerChannel>> channelClasses;

  @Autowired
  protected List<ChannelOptionItem<?>> channelOptions;
  @Autowired
  protected List<ChannelHandler> channelHandlers;
  @Autowired
  protected ChannelInitializer<?> channelInitializer;
  @Autowired
  protected SocketAddress address;

  public void start() throws Exception {
    EventLoopGroup bossGroup = new NioEventLoopGroup(this.threadCountBoss);
    EventLoopGroup workerGroup = new NioEventLoopGroup(this.threadCountWorker);

    try {
      ServerBootstrap serverBootstrap = new ServerBootstrap();
      serverBootstrap.group(bossGroup, workerGroup);

      if (null != this.channelClasses) {
        for (Class<ServerChannel> channelClass : this.channelClasses) {
          serverBootstrap.channel(channelClass);
        }
      }
      if (null != this.channelOptions) {
        for (ChannelOptionItem<?> optionItem : this.channelOptions) {
          optionItem.setOption(serverBootstrap);
        }
      }
      if (null != this.channelHandlers) {
        for (ChannelHandler channelHandler : this.channelHandlers) {
          serverBootstrap.handler(channelHandler);
        }
      }
      serverBootstrap.childHandler(this.channelInitializer);

      this.init();

      ChannelFuture f = serverBootstrap.bind(address).sync();
      this.onStart();

      f.channel().closeFuture().sync();
    } catch (Exception e) {

    } finally {
      this.onStop();
      bossGroup.shutdownGracefully();
      workerGroup.shutdownGracefully();
      this.destroy();
    }
  }

  protected abstract void init() throws Exception;
  protected abstract void onStart() throws Exception;
  protected abstract void onStop() throws Exception;
  protected abstract void destroy() throws Exception;

  public static class ChannelOptionItem<T> {
    private final ChannelOption<T> option;
    private final T value;

    public ChannelOptionItem(ChannelOption<T> option, T value) {
      this.option = option;
      this.value = value;
    }
    void setOption(ServerBootstrap serverBootstrap) {
      serverBootstrap.option(option, value);
    }
  }
}